from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw
import narwhals.selectors as ncs
import numpy as np

if TYPE_CHECKING:
    from narhwals.typing import FrameT


class CheckNumericMixin:
    """
    Mixin class with methods for numeric transformers

    """

    def classname(self) -> str:
        """Method that returns the name of the current class when called."""

        return type(self).__name__

    @nw.narwhalify
    def check_numeric_columns(self, X: FrameT) -> FrameT:
        """Helper function for checking column args are numeric for numeric transformers.

        Args:
        ----
            X: Data containing columns to check.

        """
        non_numeric_columns = list(
            set(self.columns).difference(set(X.select(ncs.numeric()).columns)),
        )
        # sort as set ordering can be inconsistent
        non_numeric_columns.sort()
        if len(non_numeric_columns) > 0:
            msg = f"{self.classname()}: The following columns are not numeric in X; {non_numeric_columns}"
            raise TypeError(msg)

        return X


class DropOriginalMixin:
    """Mixin class to validate and apply 'drop_original' argument used by various transformers.

    Transformer deletes transformer input columns depending on boolean argument.

    """

    def classname(self) -> str:
        """Method that returns the name of the current class when called."""

        return type(self).__name__

    def set_drop_original_column(self, drop_original: bool) -> None:
        """Helper method for validating 'drop_original' argument.

        Parameters
        ----------
        drop_original : bool
            boolean dictating dropping the input columns from X after checks.

        """
        # check if 'drop_original' argument is boolean
        if type(drop_original) is not bool:
            msg = f"{self.classname()}: drop_original should be bool"
            raise TypeError(msg)

        self.drop_original = drop_original

    @nw.narwhalify
    def drop_original_column(
        self,
        X: FrameT,
        drop_original: bool,
        columns: list[str] | str | None,
    ) -> FrameT:
        """Method for dropping input columns from X if drop_original set to True.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data with columns to drop.

        drop_original : bool
            boolean dictating dropping the input columns from X after checks.

        columns: list[str] | str |  None
            Object containing columns to drop

        Returns
        -------
        X : pd/pl.DataFrame
            Transformed input X with columns dropped.

        """

        if drop_original:
            X = X.drop(columns)

        return X


class NewColumnNameMixin:
    """Helper to validate and set new_column_name attribute"""

    def check_and_set_new_column_name(self, new_column_name: str) -> None:
        if not (isinstance(new_column_name, str)):
            msg = f"{self.classname()}: new_column_name should be str"
            raise TypeError(msg)

        self.new_column_name = new_column_name


class SeparatorColumnMixin:
    """Hel per to validate and set separator attribute"""

    def check_and_set_separator_column(self, separator: str) -> None:
        if not (isinstance(separator, str)):
            msg = f"{self.classname()}: separator should be str"
            raise TypeError(msg)

        self.separator = separator


class TwoColumnMixin:
    """helper to validate columns when exactly two columns are required"""

    def check_two_columns(self, columns: list[str]) -> None:
        if not (isinstance(columns, list)):
            msg = f"{self.classname()}: columns should be list"
            raise TypeError(msg)

        if len(columns) != 2:
            msg = f"{self.classname()}: This transformer works with two columns only"
            raise ValueError(msg)


class WeightColumnMixin:
    """
    Mixin class with weights functionality

    """

    def classname(self) -> str:
        """Method that returns the name of the current class when called."""
        return type(self).__name__

    @nw.narwhalify
    def check_weights_column(self, X: FrameT, weights_column: str) -> None:
        """Helper method for validating weights column in dataframe.

        Args:
        ----
            X: pandas or polars df containing weight column
            weights_column: name of weight column

        """
        # check if given weight is in columns
        if weights_column not in X.columns:
            msg = f"{self.classname()}: weight col ({weights_column}) is not present in columns of data"
            raise ValueError(msg)

        # check weight is numeric
        if weights_column not in X.select(ncs.numeric()).columns:
            msg = f"{self.classname()}: weight column must be numeric."
            raise ValueError(msg)

        # check weight is positive
        if X.select(nw.col(weights_column).min()).item() < 0:
            msg = f"{self.classname()}: weight column must be positive"
            raise ValueError(msg)

        if (
            # check weight non-None
            (X.select(nw.col(weights_column).is_null().sum()).item() != 0)
            or
            # TODO - update this section with narwhals is_nan method once this becomes
            # available
            # check weight non-NaN - polars differentiates between None and NaN
            (np.isnan(X.get_column(weights_column).to_numpy()).sum() != 0)
        ):
            msg = f"{self.classname()}: weight column must be non-null"
            raise ValueError(msg)

        # check weight not inf
        if not X.select((nw.col(weights_column).is_finite()).all()).item():
            msg = f"{self.classname()}: weight column must not contain infinite values."
            raise ValueError(msg)

        # check weight not all 0
        if X.select(nw.col(weights_column).sum()).item() == 0:
            msg = f"{self.classname()}: total sample weights are not greater than 0"
            raise ValueError(msg)

    def check_and_set_weight(self, weights_column: str) -> None:
        """Helper method that validates and assigns the specified column name to be used as the weights_column attribute.
        This function ensures that the `weights_column` parameter is either a string representing
        the column name or None. If `weights_column` is not of type str and is not None, it raises
        a TypeError.

        Parameters:
            weights_column (str or None): The name of the column to be used as weights. If None, no weights are used.

        Raises:
            TypeError: If `weights_column` is neither a string nor None.

        Returns:
            None
        """

        if weights_column is not None and not isinstance(weights_column, str):
            msg = "weights_column should be str or None"
            raise TypeError(msg)
        self.weights_column = weights_column



import datetime

import numpy as np
import polars as pl
import pytest

import tests.test_data as d
from tests.base_tests import (
    DropOriginalInitMixinTests,
    DropOriginalTransformMixinTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
    TwoColumnListInitTests,
)
from tests.dates.test_BaseGenericDateTransformer import (
    GenericDatesMixinTransformTests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.dates import DateDifferenceTransformer


class TestInit(
    TwoColumnListInitTests,
    DropOriginalInitMixinTests,
    NewColumnNameInitMixintests,
):
    """Tests for DateDifferenceTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DateDifferenceTransformer"

    def test_units_values_error(self):
        """Test that an exception is raised if the value of inits is not one of accepted_values_units."""
        with pytest.raises(
            ValueError,
            match=r"DateDifferenceTransformer: units must be one of \['D', 'h', 'm', 's'\], got y",
        ):
            DateDifferenceTransformer(
                columns=["dummy_1", "dummy_2"],
                new_column_name="dummy_3",
                units="y",
                verbose=False,
            )


def expected_df_3(library="pandas"):
    """Expected output for test_expected_output_units_D."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2000,
                3,
                19,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                10,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                12,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                1985,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "b": [
            datetime.datetime(
                2020,
                5,
                1,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                9,
                10,
                9,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "D": [
            9713.042372685186,
            7219.957627314815,
            0.0,
            31.0,
            -30.083333333333332,
            -1064.9583333333333,
            -1125.9583333333333,
            10957.0,
        ],
    }

    return dataframe_init_dispatch(df_dict, library=library)


def expected_df_4(library="pandas"):
    """Expected output for test_expected_output_units_h."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2000,
                3,
                19,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                10,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                12,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                1985,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "b": [
            datetime.datetime(
                2020,
                5,
                1,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                9,
                10,
                9,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "h": [
            233113.01694444445,
            173278.98305555555,
            0.0,
            744.0,
            -722.0,
            -25559.0,
            -27023.0,
            262968.0,
        ],
    }

    return dataframe_init_dispatch(df_dict, library=library)


def expected_df_5(library="pandas"):
    """Expected output for test_expected_output_units_m."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2000,
                3,
                19,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                10,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                12,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                1985,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "b": [
            datetime.datetime(
                2020,
                5,
                1,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                9,
                10,
                9,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "m": [
            13986781.016666668,
            10396738.983333332,
            0.0,
            44640.0,
            -43320.0,
            -1533540.0,
            -1621380.0,
            15778080.0,
        ],
    }

    return dataframe_init_dispatch(df_dict, library=library)


def expected_df_6(library="pandas"):
    """Expected output for test_expected_output_units_s."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2000,
                3,
                19,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                10,
                10,
                10,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                12,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                1985,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "b": [
            datetime.datetime(
                2020,
                5,
                1,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2018,
                9,
                10,
                9,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                11,
                10,
                12,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "s": [
            839206861.0,
            623804339.0,
            0.0,
            2678400.0,
            -2599200.0,
            -92012400.0,
            -97282800.0,
            946684800.0,
        ],
    }

    return dataframe_init_dispatch(df_dict, library=library)


def expected_df_7(library="pandas"):
    """Expected output for test_expected_output_nulls."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            np.nan if library == "pandas" else None,
            datetime.datetime(
                1985,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "b": [
            np.nan if library == "pandas" else None,
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "D": [
            np.nan if library == "pandas" else None,
            np.nan if library == "pandas" else None,
            10957.0,
        ],
    }

    return dataframe_init_dispatch(df_dict, library=library)


def create_datediff_test_nulls_df(library="pandas"):
    """Create DataFrame with nulls for DateDifferenceTransformer tests."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            np.nan if library == "pandas" else None,
            datetime.datetime(
                1985,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "b": [
            np.nan if library == "pandas" else None,
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            datetime.datetime(
                2015,
                7,
                23,
                11,
                59,
                59,
                tzinfo=datetime.timezone.utc,
            ),
        ],
    }

    return dataframe_init_dispatch(df_dict, library=library)


def expected_df_8(library="pandas"):
    """Expected output for test_expected_output_nulls2."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            np.nan if library == "pandas" else None,
        ],
        "b": [
            np.nan if library == "pandas" else None,
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
        ],
        "D": [
            np.nan if library == "pandas" else None,
            np.nan if library == "pandas" else None,
        ],
    }
    if library == "polars":
        expected = dataframe_init_dispatch(df_dict, library=library)
        # change the D column to float as in the transformer it is a calculated field and this is automatically set up as float.
        return expected.cast({"D": pl.Float64})
    return dataframe_init_dispatch(df_dict, library=library)


def create_datediff_test_nulls_df2(library="pandas"):
    """Create DataFrame with nulls only for DateDifferenceTransformer tests."""

    df_dict = {
        "a": [
            datetime.datetime(
                1993,
                9,
                27,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
            np.nan if library == "pandas" else None,
        ],
        "b": [
            np.nan if library == "pandas" else None,
            datetime.datetime(
                2019,
                12,
                25,
                11,
                58,
                58,
                tzinfo=datetime.timezone.utc,
            ),
        ],
    }

    return dataframe_init_dispatch(df_dict, library=library)


class TestTransform(
    GenericTransformTests,
    GenericDatesMixinTransformTests,
    DropOriginalTransformMixinTests,
):
    """Tests for DateDifferenceTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DateDifferenceTransformer"

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_datediff_test_df(library="pandas"),
                expected_df_3(library="pandas"),
            ),
            (
                d.create_datediff_test_df(library="polars"),
                expected_df_3(library="polars"),
            ),
        ],
    )
    def test_expected_output_units_D(self, df, expected):
        """Test that the output is expected from transform, when units is D.

        This tests positive month gaps, negative month gaps, and missing values.

        """
        x = DateDifferenceTransformer(
            columns=["a", "b"],
            new_column_name="D",
            units="D",
            verbose=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(expected, df_transformed)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_datediff_test_df(library="pandas"),
                expected_df_4(library="pandas"),
            ),
            (
                d.create_datediff_test_df(library="polars"),
                expected_df_4(library="polars"),
            ),
        ],
    )
    def test_expected_output_units_h(self, df, expected):
        """Test that the output is expected from transform, when units is h.

        This tests positive month gaps, negative month gaps, and missing values.

        """
        x = DateDifferenceTransformer(
            columns=["a", "b"],
            new_column_name="h",
            units="h",
            verbose=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(expected, df_transformed)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_datediff_test_df(library="pandas"),
                expected_df_5(library="pandas"),
            ),
            (
                d.create_datediff_test_df(library="polars"),
                expected_df_5(library="polars"),
            ),
        ],
    )
    def test_expected_output_units_m(self, df, expected):
        """Test that the output is expected from transform, when units is m.

        This tests positive month gaps, negative month gaps, and missing values.

        """
        x = DateDifferenceTransformer(
            columns=["a", "b"],
            new_column_name="m",
            units="m",
            verbose=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_datediff_test_df(library="pandas"),
                expected_df_6(library="pandas"),
            ),
            (
                d.create_datediff_test_df(library="polars"),
                expected_df_6(library="polars"),
            ),
        ],
    )
    def test_expected_output_units_s(self, df, expected):
        """Test that the output is expected from transform, when units is s.

        This tests positive month gaps, negative month gaps, and missing values.

        """
        x = DateDifferenceTransformer(
            columns=["a", "b"],
            new_column_name="s",
            units="s",
            verbose=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(expected, df_transformed)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                create_datediff_test_nulls_df(library="pandas"),
                expected_df_7(library="pandas"),
            ),
            (
                create_datediff_test_nulls_df(library="polars"),
                expected_df_7(library="polars"),
            ),
        ],
    )
    def test_expected_output_nulls(self, df, expected):
        """Test that the output is expected from transform, when columns have nulls."""
        x = DateDifferenceTransformer(
            columns=["a", "b"],
            new_column_name="D",
            units="D",
            verbose=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                create_datediff_test_nulls_df2(library="pandas"),
                expected_df_8(library="pandas"),
            ),
            (
                create_datediff_test_nulls_df2(library="polars"),
                expected_df_8(library="polars"),
            ),
        ],
    )
    def test_expected_output_nulls2(self, df, expected):
        """Test that the output is expected from transform, when columns are nulls."""
        x = DateDifferenceTransformer(
            columns=["a", "b"],
            new_column_name="D",
            units="D",
            verbose=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DateDifferenceTransformer"



"""This module contains transformers that other transformers in the package inherit
from. These transformers contain key checks to be applied in all cases.
"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Union

import narwhals as nw
import pandas as pd
from beartype import beartype
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from tubular.mixins import DropOriginalMixin

if TYPE_CHECKING:
    from narwhals.typing import FrameT

pd.options.mode.copy_on_write = True


class BaseTransformer(TransformerMixin, BaseEstimator):
    """Base tranformer class which all other transformers in the package inherit from.

    Provides fit and transform methods (required by sklearn transformers), simple input checking
    and functionality to copy X prior to transform.

    Parameters
    ----------
    columns : None or list or str
        Columns to apply the transformer to. If a str is passed this is put into a list. Value passed
        in columns is saved in the columns attribute on the object.

    copy : bool, default = True
        Should X be copied before tansforms are applied? Copy argument no longer used and will be deprecated in a future release

    verbose : bool, default = False
        Should statements be printed when methods are run?

    Attributes
    ----------
    columns : list
        Either a list of str values giving which columns in a input pandas.DataFrame the transformer
        will be applied to.

    copy : bool
        Should X be copied before tansforms are applied? Copy argument no longer used and will be deprecated in a future release

    verbose : bool
        Print statements to show which methods are being run or not.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    def classname(self) -> str:
        """Method that returns the name of the current class when called."""
        return type(self).__name__

    @beartype
    def __init__(
        self,
        columns: Union[list[str], str],
        copy: Union[bool, None] = None,
        verbose: bool = False,
    ) -> None:
        if copy is not None:
            warnings.warn(
                "copy argument no longer used and will be deprecated in a future release",
                DeprecationWarning,
                stacklevel=2,
            )

        self.verbose = verbose

        if self.verbose:
            print("BaseTransformer.__init__() called")

        # make sure columns is a single str or list of strs
        if isinstance(columns, str):
            self.columns = [columns]

        elif isinstance(columns, list):
            if not len(columns) > 0:
                msg = f"{self.classname()}: columns has no values"
                raise ValueError(msg)

            self.columns = columns

        self.copy = copy

    @nw.narwhalify
    def fit(self, X: FrameT, y: nw.Series | None = None) -> BaseTransformer:
        """Base transformer fit method, checks X and y types. Currently only pandas DataFrames are allowed for X
        and DataFrames or Series for y.

        Fit calls the columns_check method which will check that the columns attribute is set and all values are present in X

        Parameters
        ----------
        X : pd.DataFrame
            Data to fit the transformer on.

        y : None or pd.DataFrame or pd.Series, default = None
            Optional argument only required for the transformer to work with sklearn pipelines.

        """
        if self.verbose:
            print("BaseTransformer.fit() called")

        self.columns_check(X)

        if not X.shape[0] > 0:
            msg = f"{self.classname()}: X has no rows; {X.shape}"
            raise ValueError(msg)

        if y is not None:
            if not isinstance(y, nw.Series):
                msg = f"{self.classname()}: unexpected type for y, should be a polars or pandas Series"
                raise TypeError(msg)

            if not y.shape[0] > 0:
                msg = f"{self.classname()}: y is empty; {y.shape}"
                raise ValueError(msg)

        return self

    @nw.narwhalify
    def _combine_X_y(self, X: FrameT, y: nw.Series) -> FrameT:
        """Combine X and y by adding a new column with the values of y to a copy of X.

        The new column response column will be called `_temporary_response`.

        This method can be used by transformers that need to use the response, y, together
        with the explanatory variables, X, in their `fit` methods.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data containing explanatory variables.

        y : pd/pl.Series
            Response variable.

        """
        if not isinstance(X, (nw.DataFrame, nw.LazyFrame)):
            msg = f"{self.classname()}: X should be a polars or pandas DataFrame/LazyFrame"
            raise TypeError(msg)

        if not isinstance(y, nw.Series):
            msg = f"{self.classname()}: y should be a polars or pandas Series"
            raise TypeError(msg)

        if X.shape[0] != y.shape[0]:
            msg = f"{self.classname()}: X and y have different numbers of rows ({X.shape[0]} vs {y.shape[0]})"
            raise ValueError(msg)

        return X.with_columns(_temporary_response=y)

    @nw.narwhalify
    def transform(self, X: FrameT) -> FrameT:
        """Base transformer transform method; checks X type (pandas/polars DataFrame only) and copies data if requested.

        Transform calls the columns_check method which will check columns in columns attribute are in X.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to transform with the transformer.

        Returns
        -------
        X : pd/pl.DataFrame
            Input X, copied if specified by user.

        """
        self.columns_check(X)

        if self.verbose:
            print("BaseTransformer.transform() called")

        # to prevent overwriting original dataframe
        X_view = X.clone()

        if not X.shape[0] > 0:
            msg = f"{self.classname()}: X has no rows; {X.shape}"
            raise ValueError(msg)

        return X_view

    def check_is_fitted(self, attribute: str) -> None:
        """Check if particular attributes are on the object. This is useful to do before running transform to avoid
        trying to transform data without first running the fit method.

        Wrapper for utils.validation.check_is_fitted function.

        Parameters
        ----------
        attributes : List
            List of str values giving names of attribute to check exist on self.

        """
        check_is_fitted(self, attribute)

    @nw.narwhalify
    def columns_check(self, X: FrameT) -> None:
        """Method to check that the columns attribute is set and all values are present in X.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to check columns are in.

        """
        if not isinstance(X, (nw.DataFrame, nw.LazyFrame)):
            msg = f"{self.classname()}: X should be a polars or pandas DataFrame/LazyFrame"
            raise TypeError(msg)

        if not isinstance(self.columns, list):
            msg = f"{self.classname()}: self.columns should be a list"
            raise TypeError(msg)

        for c in self.columns:
            if c not in X.columns:
                raise ValueError(f"{self.classname()}: variable " + c + " is not in X")


class DataFrameMethodTransformer(DropOriginalMixin, BaseTransformer):

    """Tranformer that applies a pandas.DataFrame method.

    Transformer assigns the output of the method to a new column or columns. It is possible to
    supply other key word arguments to the transform method, which will be passed to the
    pandas.DataFrame method being called.

    Be aware it is possible to supply incompatible arguments to init that will only be
    identified when transform is run. This is because there are many combinations of method, input
    and output sizes. Additionally some methods may only work as expected when called in
    transform with specific key word arguments.

    Parameters
    ----------
    new_column_names : str or list of str
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    pd_method_name : str
        The name of the pandas.DataFrame method to call.

    columns : None or list or str
        Columns to apply the transformer to. If a str is passed this is put into a list. Value passed
        in columns is saved in the columns attribute on the object. Note this has no default value so
        the user has to specify the columns when initialising the transformer. This is avoid likely
        when the user forget to set columns, in this case all columns would be picked up when super
        transform runs.

    pd_method_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.DataFrame method when it is called.

    drop_original : bool, default = False
        Should original columns be dropped?

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    new_column_names : str or list of str
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    pd_method_name : str
        The name of the pandas.DataFrame method to call.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = False

    def __init__(
        self,
        new_column_names: list[str] | str,
        pd_method_name: str,
        columns: list[str] | str | None,
        pd_method_kwargs: dict[str, object] | None = None,
        drop_original: bool = False,
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        if type(new_column_names) is list:
            for i, item in enumerate(new_column_names):
                if type(item) is not str:
                    msg = f"{self.classname()}: if new_column_names is a list, all elements must be strings but got {type(item)} in position {i}"
                    raise TypeError(msg)

        elif type(new_column_names) is not str:
            msg = f"{self.classname()}: unexpected type ({type(new_column_names)}) for new_column_names, must be str or list of strings"
            raise TypeError(msg)

        if type(pd_method_name) is not str:
            msg = f"{self.classname()}: unexpected type ({type(pd_method_name)}) for pd_method_name, expecting str"
            raise TypeError(msg)

        if pd_method_kwargs is None:
            pd_method_kwargs = {}
        else:
            if type(pd_method_kwargs) is not dict:
                msg = f"{self.classname()}: pd_method_kwargs should be a dict but got type {type(pd_method_kwargs)}"
                raise TypeError(msg)

            for i, k in enumerate(pd_method_kwargs.keys()):
                if type(k) is not str:
                    msg = f"{self.classname()}: unexpected type ({type(k)}) for pd_method_kwargs key in position {i}, must be str"
                    raise TypeError(msg)

        self.new_column_names = new_column_names
        self.pd_method_name = pd_method_name
        self.pd_method_kwargs = pd_method_kwargs

        DropOriginalMixin.set_drop_original_column(self, drop_original)

        try:
            df = pd.DataFrame()
            getattr(df, pd_method_name)

        except Exception as err:
            msg = f'{self.classname()}: error accessing "{pd_method_name}" method on pd.DataFrame object - pd_method_name should be a pd.DataFrame method'
            raise AttributeError(msg) from err

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform input pandas DataFrame (X) using the given pandas.DataFrame method and assign the output
        back to column or columns in X.

        Any keyword arguments set in the pd_method_kwargs attribute are passed onto the pandas DataFrame method when calling it.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column or columns (self.new_column_names) added. These contain the output of
            running the pandas DataFrame method.

        """
        X = super().transform(X)

        X[self.new_column_names] = getattr(X[self.columns], self.pd_method_name)(
            **self.pd_method_kwargs,
        )

        # Drop original columns if self.drop_original is True
        return DropOriginalMixin.drop_original_column(
            self,
            X,
            self.drop_original,
            self.columns,
        )



import re

import narwhals as nw
import pandas as pd
import polars as pl
import pytest

import tests.test_data as d
import tests.utils as u
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tests.imputers.test_BaseImputer import GenericImputerTransformTests
from tubular.imputers import ArbitraryImputer


def impute_df_with_several_types(library="pandas"):
    """
    Fixture that returns a DataFrame with columns suitable for downcasting
    for both pandas and polars.
    """
    data = {
        "a": ["a", "b", "c", "d", None],
        "b": [1.0, 2.0, 3.0, 4.0, None],
        "c": [True, False, False, None, True],
    }

    return u.dataframe_init_dispatch(data, library)


class TestInit(ColumnStrListInitTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ArbitraryImputer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ArbitraryImputer"


class TestTransform(GenericImputerTransformTests, GenericTransformTests):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ArbitraryImputer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("column", "col_type", "impute_value"),
        [
            ("a", "String", 1),
            ("a", "Categorical", True),
            ("b", "Float32", "bla"),
            ("c", "Boolean", 500),
        ],
    )
    def test_type_mismatch_errors(
        self,
        column,
        col_type,
        impute_value,
        library,
    ):
        """Test that dtypes are preserved after imputation."""

        df = impute_df_with_several_types(library=library)

        df = nw.from_native(df)

        df = df.with_columns(
            nw.col(column).cast(getattr(nw, col_type)),
        )

        df = nw.to_native(df)

        transformer = ArbitraryImputer(impute_value=impute_value, columns=[column])

        if col_type in ["Categorical", "String"]:
            msg_required_impute_value_type = "str"
            msg_col_type = "Categorical or String"

        elif col_type == "Boolean":
            msg_required_impute_value_type = "bool"
            msg_col_type = "Boolean"

        else:
            msg_required_impute_value_type = "numeric"
            msg_col_type = "Numeric"

        msg = rf"""
                {self.transformer_name}: Attempting to impute non-{msg_required_impute_value_type} value {transformer.impute_value} into
                {msg_col_type} type columns, this is not type safe,
                please use {msg_required_impute_value_type} impute_value for these columns
                \(this may require separate ArbitraryImputer instances for different column types\)
                """

        with pytest.raises(
            TypeError,
            match=msg,
        ):
            transformer.transform(df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("column", "col_type", "impute_value", "expected_values"),
        [
            ("a", "String", "z", ["a", "b", "c", "d", "z"]),
            ("a", "Categorical", "z", ["a", "b", "c", "d", "z"]),
            ("b", "Float32", 1, [1.0, 2.0, 3.0, 4.0, 1.0]),
        ],
    )
    def test_impute_value_preserve_dtype(
        self,
        column,
        col_type,
        impute_value,
        expected_values,
        library,
    ):
        """Test that dtypes are preserved after imputation."""

        df = impute_df_with_several_types(library=library)

        df_nw = nw.from_native(df)

        # or just downcast easier types (String comes in correct type so leave)
        if col_type in ["Float32", "Categorical"]:
            df_nw = df_nw.with_columns(
                nw.col(column).cast(getattr(nw, col_type)),
            )

        transformer = ArbitraryImputer(impute_value=impute_value, columns=[column])
        df_transformed_native = transformer.transform(df_nw.to_native())

        df_transformed_nw = nw.from_native(df_transformed_native)

        expected_dtype = df_nw[column].dtype
        actual_dtype = df_transformed_nw[column].dtype
        assert (
            actual_dtype == expected_dtype
        ), f"{self.transformer_name}: dtype changed unexpectedly in transform, expected {expected_dtype} but got {actual_dtype}"

        # also check full df against expectation
        expected = df_nw.clone()
        expected = expected.with_columns(
            nw.new_series(name=column, values=expected_values, backend=library).cast(
                getattr(nw, col_type),
            ),
        )

        u.assert_frame_equal_dispatch(expected.to_native(), df_transformed_native)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("input_col", "expected_dtype", "impute_value", "expected_values"),
        [
            ([None, None], "String", "a", ["a", "a"]),
            ([True, False, None], "Boolean", True, [True, False, True]),
        ],
    )
    def test_edge_cases(
        self,
        input_col,
        expected_dtype,
        impute_value,
        expected_values,
        library,
    ):
        """Test handling for some edge cases:
        - pandas object type
        - all null column
        """

        column = "a"
        df_dict = {"a": input_col}

        df = u.dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

        df_nw = nw.from_native(df)

        transformer = ArbitraryImputer(impute_value=impute_value, columns=[column])

        # for pandas, the all null column is inferred as string type
        # for polars, it is Unknown type, which triggers a warning
        if library == "polars" and input_col == [None, None]:
            with pytest.warns(
                UserWarning,
                match=f"{self.transformer_name}: X contains all null columns {str({column})}, types for these columns will be inferred as {type(transformer.impute_value)}",
            ):
                df_transformed_native = transformer.transform(df_nw.to_native())

        else:
            df_transformed_native = transformer.transform(df_nw.to_native())

        df_transformed_nw = nw.from_native(df_transformed_native)

        actual_dtype = str(df_transformed_nw[column].dtype)

        assert (
            actual_dtype == expected_dtype
        ), f"{self.transformer_name}: dtype changed unexpectedly in transform, expected {expected_dtype} but got {actual_dtype}"

        # also check full df against expectation
        expected = df_nw.clone()
        expected = expected.with_columns(
            nw.new_series(name=column, values=expected_values, backend=library).cast(
                getattr(nw, expected_dtype),
            ),
        )

        u.assert_frame_equal_dispatch(expected.to_native(), df_transformed_native)

    @pytest.mark.parametrize(
        ("impute_value", "impute_val_type"),
        [
            (1, "Int32"),
            ("a", "String"),
            (True, "Boolean"),
        ],
    )
    def test_polars_unknown_type_output(self, impute_value, impute_val_type):
        """Test handling of polars Unknown type column (output type should be inferred from impute_value)"""

        column = "a"
        values = [None, None]
        df_dict = {"a": values}

        df = pl.DataFrame(df_dict)

        df_nw = nw.from_native(df)

        transformer = ArbitraryImputer(impute_value=impute_value, columns=[column])

        df_transformed_native = transformer.transform(df_nw.to_native())

        df_transformed_nw = nw.from_native(df_transformed_native)

        actual_dtype = str(df_transformed_nw[column].dtype)

        assert (
            actual_dtype == impute_val_type
        ), f"{self.transformer_name}: dtype changed unexpectedly in transform, expected {impute_val_type} but got {actual_dtype}"

        # also check full df against expectation
        expected = df_nw.clone()
        expected = expected.with_columns(
            nw.new_series(
                name=column,
                values=[impute_value, impute_value],
                backend="polars",
            ).cast(getattr(nw, impute_val_type)),
        )

        u.assert_frame_equal_dispatch(expected.to_native(), df_transformed_native)

    @pytest.mark.parametrize(
        ("library", "expected_df_4", "impute_values_dict"),
        [
            ("pandas", "pandas", {"b": "z", "c": "z"}),
            ("polars", "polars", {"b": "z", "c": "z"}),
        ],
        indirect=["expected_df_4"],
    )
    def test_expected_output_4(
        self,
        library,
        expected_df_4,
        initialized_transformers,
        impute_values_dict,
    ):
        """Test that transform is giving the expected output when applied to object and categorical columns
        (when we're imputing with a new categorical level, which is only possible for arbitrary imputer).
        """
        # Create the DataFrame using the library parameter
        df2 = d.create_df_2(library=library)

        # Initialize the transformer
        transformer = initialized_transformers[self.transformer_name]

        transformer.impute_values_ = impute_values_dict
        transformer.impute_value = "z"
        transformer.columns = ["b", "c"]

        # Transform the DataFrame
        df_transformed = transformer.transform(df2)

        # Check whole dataframes
        u.assert_frame_equal_dispatch(
            df_transformed,
            expected_df_4,
        )
        df2 = nw.from_native(df2)
        expected_df_4 = nw.from_native(expected_df_4)

        # Check outcomes for single rows
        # turn off type change errors to avoid having to type the single rows
        transformer.error_on_type_change = False
        for i in range(len(df2)):
            df_transformed_row = transformer.transform(df2[[i]].to_native())
            df_expected_row = expected_df_4[[i]].to_native()

            u.assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        "input_values",
        [
            [["a", "b"], ["c", "d"]],
            [{"a": 1}, {"b": 4}],
        ],
    )
    def test_weird_dtype_errors(
        self,
        input_values,
        library,
    ):
        """Test that unexpected dtypes will hit error"""

        column = "a"
        df_dict = {column: input_values}

        # because of weird types, initialise manually
        df = pd.DataFrame(df_dict) if library == "pandas" else pl.DataFrame(df_dict)

        transformer = ArbitraryImputer(impute_value=1, columns=[column])

        msg = re.escape(
            f"""
                {self.transformer_name}: transformer can only handle Float/Int/Boolean/String/Categorical/Unknown type columns
                but got columns with types {nw.from_native(df).select(nw.col(column)).schema}
                """,
        )

        with pytest.raises(
            TypeError,
            match=msg,
        ):
            transformer.transform(df)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwrite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "ArbitraryImputer"



import pytest

import tests.test_data as d
import tests.utils as u
from tests.base_tests import DropOriginalInitMixinTests, DropOriginalTransformMixinTests
from tests.numeric.test_BaseNumericTransformer import (
    BaseNumericTransformerFitTests,
    BaseNumericTransformerInitTests,
    BaseNumericTransformerTransformTests,
)
from tubular.numeric import OneDKmeansTransformer


class TestInit(
    BaseNumericTransformerInitTests,
    DropOriginalInitMixinTests,
):
    """Tests for OneDKmeansTransformer.init()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OneDKmeansTransformer"

    # Those checks are replaced with @beartype so they can be skipped
    def test_drop_column_arg_errors(self):
        pass

    def test_clone(self):
        pass

    def test_acolumns_type_error_if_not_str_or_len1_list(self):
        """Test that an exception is raised if kmeans_kwargs is not a dict."""
        with pytest.raises(
            TypeError,
            match=r"""OneDKmeansTransformer: columns arg should be a single str or a list length 1 giving the column to group.""",
        ):
            OneDKmeansTransformer(columns=["b", "c"], new_column_name="a")


class TestFit(BaseNumericTransformerFitTests):
    """Tests for OneDKmeansTransformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OneDKmeansTransformer"

    # This is dealt by @deartype
    def test_unexpected_kwarg_error(self):
        pass

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("df_generator"),
        [
            (d.create_df_9),  # int with None
            (d.create_bool_and_float_df),  # float with np.nan
            (d.create_df_with_none_and_nan_cols),  # all np.nan
        ],
    )
    def test_x_nans_value_type_error(
        self,
        library,
        df_generator,
    ):
        """Test that an exception is raised if X contains Nan or None."""
        with pytest.raises(
            ValueError,
            match=r"OneDKmeansTransformer: X should not contain missing values.",
        ):
            df = df_generator(library=library)
            OneDKmeansTransformer(
                new_column_name="b",
                columns="a",
            ).fit(X=df)


# Create test data
def create_numeric_df_1(library="pandas"):
    """Example with numeric dataframe."""

    df_dict = {
        "a": [4, 5, 4, 5, 2, 1, 3, 2, 1, 5],
        "b": [43, 77, 61, 29, 84, 29, 24, 40, 84, 96],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def expected_numeric_df_1(library="pandas"):
    """Example with numeric dataframe."""

    df_dict = {
        "a": [4, 5, 4, 5, 2, 1, 3, 2, 1, 5],
        "b": [43, 77, 61, 29, 84, 29, 24, 40, 84, 96],
        "new": [0, 1, 0, 0, 1, 0, 0, 0, 1, 1],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def expected_numeric_df_1_drop(library="pandas"):
    """Example with numeric dataframe."""

    df_dict = {
        "a": [4, 5, 4, 5, 2, 1, 3, 2, 1, 5],
        "new": [0, 1, 0, 0, 1, 0, 0, 0, 1, 1],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def create_numeric_df_2(library="pandas"):
    """Example with numeric dataframe."""

    df_dict = {
        "a": [4, 5, 4, 5, 2, 1, 3, 2, 1, 5, 10, 12, 4, 16, 17],
        "b": [43, -77, -61, 29, 84, 29, -24, 40, 84, -96, 10, -4, 15, -12, 15],
        "c": [
            "a",
            "b",
            "a",
            "b",
            "a",
            "b",
            "b",
            "a",
            "c",
            "b",
            "a",
            "c",
            "a",
            "c",
            "a",
        ],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def expected_numeric_df_2(library="pandas"):
    """Example with numeric dataframe."""

    df_dict = {
        "a": [4, 5, 4, 5, 2, 1, 3, 2, 1, 5, 10, 12, 4, 16, 17],
        "b": [43, -77, -61, 29, 84, 29, -24, 40, 84, -96, 10, -4, 15, -12, 15],
        "c": [
            "a",
            "b",
            "a",
            "b",
            "a",
            "b",
            "b",
            "a",
            "c",
            "b",
            "a",
            "c",
            "a",
            "c",
            "a",
        ],
        "new": [3, 0, 0, 3, 4, 3, 1, 3, 4, 0, 2, 1, 2, 1, 2],
    }

    return u.dataframe_init_dispatch(df_dict, library)


def expected_numeric_df_2_drop(library="pandas"):
    """Example with numeric dataframe."""

    df_dict = {
        "a": [4, 5, 4, 5, 2, 1, 3, 2, 1, 5, 10, 12, 4, 16, 17],
        "c": [
            "a",
            "b",
            "a",
            "b",
            "a",
            "b",
            "b",
            "a",
            "c",
            "b",
            "a",
            "c",
            "a",
            "c",
            "a",
        ],
        "new": [3, 0, 0, 3, 4, 3, 1, 3, 4, 0, 2, 1, 2, 1, 2],
    }

    return u.dataframe_init_dispatch(df_dict, library)


class TestTransform(
    BaseNumericTransformerTransformTests,
    DropOriginalTransformMixinTests,
):
    """Tests for OneDKmeansTransformer.transform()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OneDKmeansTransformer"

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                create_numeric_df_1(library="pandas"),
                expected_numeric_df_1(library="pandas"),
            ),
            (
                create_numeric_df_1(library="polars"),
                expected_numeric_df_1(library="polars"),
            ),
        ],
    )
    def test_expected_output_without_drop(self, df, expected):
        """Test that the output is expected from transform, when there are no negative numbers and dont drop original"""
        x = OneDKmeansTransformer(
            columns="b",
            n_clusters=2,
            new_column_name="new",
            drop_original=False,
            kmeans_kwargs={"random_state": 42},
        ).fit(df)

        df_transformed = x.transform(df)

        u.assert_frame_equal_dispatch(expected, df_transformed)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                create_numeric_df_1(library="pandas"),
                expected_numeric_df_1_drop(library="pandas"),
            ),
            (
                create_numeric_df_1(library="polars"),
                expected_numeric_df_1_drop(library="polars"),
            ),
        ],
    )
    def test_expected_output_with_drop(self, df, expected):
        """Test that the output is expected from transform, when there are no negative numbers and drop original"""
        x = OneDKmeansTransformer(
            columns="b",
            n_clusters=2,
            new_column_name="new",
            drop_original=True,
            kmeans_kwargs={"random_state": 42},
        ).fit(df)

        df_transformed = x.transform(df)

        u.assert_frame_equal_dispatch(expected, df_transformed)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                create_numeric_df_2(library="pandas"),
                expected_numeric_df_2(library="pandas"),
            ),
            (
                create_numeric_df_2(library="polars"),
                expected_numeric_df_2(library="polars"),
            ),
        ],
    )
    def test_expected_output_without_drop_negatives(self, df, expected):
        """Test that the output is expected from transform, when there are negative numbers and dont drop original"""
        x = OneDKmeansTransformer(
            columns="b",
            n_clusters=5,
            new_column_name="new",
            drop_original=False,
            kmeans_kwargs={"random_state": 42},
        ).fit(df)

        df_transformed = x.transform(df)

        u.assert_frame_equal_dispatch(expected, df_transformed)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                create_numeric_df_2(library="pandas"),
                expected_numeric_df_2_drop(library="pandas"),
            ),
            (
                create_numeric_df_2(library="polars"),
                expected_numeric_df_2_drop(library="polars"),
            ),
        ],
    )
    def test_expected_output_with_drop_negatives(self, df, expected):
        """Test that the output is expected from transform, when there are negative numbers and drop original"""
        x = OneDKmeansTransformer(
            columns="b",
            n_clusters=5,
            new_column_name="new",
            drop_original=True,
            kmeans_kwargs={"random_state": 42},
        ).fit(df)

        df_transformed = x.transform(df)

        u.assert_frame_equal_dispatch(expected, df_transformed)



import re

import narwhals as nw
import polars as pl
import pytest

import tests.test_data as d
from tests.base_tests import GenericFitTests, GenericInitTests, GenericTransformTests
from tests.utils import dataframe_init_dispatch


class BaseNumericTransformerInitTests(GenericInitTests):
    """
    Tests for BaseNumericTransformer.init().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """


class BaseNumericTransformerFitTests(GenericFitTests):
    """
    Tests for BaseNumericTransformer.fit().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("df_generator", "bad_cols"),
        [
            (d.create_df_2, ["b"]),  # str
            (d.create_is_between_dates_df_1, ["a"]),  # datetime
            (d.create_bool_and_float_df, ["b"]),  # bool
            (d.create_df_with_none_and_nan_cols, ["b"]),  # None
        ],
    )
    def test_non_numeric_exception_raised(
        self,
        initialized_transformers,
        df_generator,
        bad_cols,
        library,
    ):
        """Test an exception is raised if self.columns are non-numeric in X."""
        df = df_generator(library=library)

        x = initialized_transformers[self.transformer_name]
        x.columns = bad_cols

        # if transformer is not polars compatible, skip polars test
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        # add in 'target column' for fit
        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df).__name__
        df = df.with_columns(
            nw.new_series(
                name="c",
                values=[1] * len(df),
                backend=native_backend,
            ),
        ).to_native()

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"{self.transformer_name}: The following columns are not numeric in X; {bad_cols}",
            ),
        ):
            x.fit(df, df["c"])

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("df_generator", "cols"),
        [
            (d.create_df_2, ["a"]),  # int
            (d.create_bool_and_float_df, ["a"]),  # float
            (d.create_df_with_none_and_nan_cols, ["a"]),  # nan
        ],
    )
    def test_numeric_passes(
        self,
        initialized_transformers,
        df_generator,
        cols,
        library,
    ):
        """Test check passes if self.columns numeric in X."""
        df = df_generator(library=library)

        x = initialized_transformers[self.transformer_name]
        x.columns = cols

        # if transformer is not polars compatible, skip polars test
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        # add in 'target column' for fit
        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df).__name__

        # Add this as OneDKmeansTransformer does not accept missing values:
        if self.transformer_name == "OneDKmeansTransformer":
            if native_backend == "polars":
                df = df.with_columns(
                    nw.when(
                        nw.col(cols[0]).is_nan(),
                    )
                    .then(
                        0,
                    )
                    .otherwise(
                        cols[0],
                    )
                    .alias(cols[0]),
                )
            df = df.with_columns(nw.col(cols[0]).fill_null(0))
            # Add this as samples are less than the 8 default clusters
            x.n_clusters = 2

        df = df.with_columns(
            nw.new_series(
                name="c",
                values=[1] * len(df),
                backend=native_backend,
            ),
        ).to_native()

        x.fit(df, df["c"])


class BaseNumericTransformerTransformTests(
    GenericTransformTests,
):
    """
    Tests for the transform method on BaseNumericTransformer.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("df_generator", "bad_cols"),
        [
            (d.create_df_2, ["b"]),  # str
            (d.create_is_between_dates_df_1, ["a"]),  # datetime
            (d.create_bool_and_float_df, ["b"]),  # bool
            (d.create_df_with_none_and_nan_cols, ["b"]),  # None
        ],
    )
    def test_non_numeric_exception_raised(
        self,
        initialized_transformers,
        df_generator,
        bad_cols,
        library,
    ):
        """Test an exception is raised if self.columns are non-numeric in X."""
        df = df_generator(library=library)

        x = initialized_transformers[self.transformer_name]
        x.columns = bad_cols

        # if transformer is not polars compatible, skip polars test
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        # add in 'target column' for and additional numeric column fit
        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df).__name__

        # Add this as samples dont have enough volumn for 8 default clusters
        if self.transformer_name == "OneDKmeansTransformer":
            # Add this as samples are less than the 8 default clusters
            x.n_clusters = 2

        df = df.with_columns(
            nw.new_series(
                name="c",
                values=[1] * len(df),
                backend=native_backend,
            ),
        ).to_native()

        # if the transformer fits, run a working fit before transform
        if x.FITS:
            # create numeric df to fit on
            df_dict = {col: df["c"] for col in [*x.columns, "c"]}
            numeric_df = dataframe_init_dispatch(
                dataframe_dict=df_dict,
                library=library,
            )
            x.fit(numeric_df, numeric_df["c"])

        with pytest.raises(
            TypeError,
            match=re.escape(
                rf"{self.transformer_name}: The following columns are not numeric in X; {x.columns}",
            ),
        ):
            x.transform(df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("df_generator"),
        [
            d.create_df_2,  # int
            d.create_bool_and_float_df,  # float
            d.create_df_with_none_and_nan_cols,  # nan
        ],
    )
    def test_numeric_passes(self, initialized_transformers, df_generator, library):
        """Test check passes if self.columns numeric in X."""
        df = df_generator(library=library)

        x = initialized_transformers[self.transformer_name]
        x.columns = ["a", "b"]

        # if transformer is not polars compatible, skip polars test
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        # add in 'target column' for and additional numeric column fit
        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df).__name__

        # Add this as OneDKmeansTransformer does not accept missing values:
        if self.transformer_name == "OneDKmeansTransformer":
            if native_backend == "polars":
                df = df.with_columns(
                    nw.when(
                        nw.col("a").is_nan(),
                    )
                    .then(
                        0,
                    )
                    .otherwise(
                        nw.col("a"),
                    )
                    .alias("a"),
                )
            df = df.with_columns(nw.col("a").fill_null(0))
            # Add this as samples are less than the 8 default clusters
            x.n_clusters = 2

        df = df.with_columns(
            nw.new_series(
                name="c",
                values=[1] * len(df),
                backend=native_backend,
            ),
            nw.new_series(
                name="b",
                values=[1] * len(df),
                backend=native_backend,
            ),
        ).to_native()

        if x.FITS:
            # create numeric df to fit on
            df_dict = {col: df["c"] for col in [*x.columns, "c"]}
            numeric_df = dataframe_init_dispatch(
                dataframe_dict=df_dict,
                library=library,
            )
            x.fit(numeric_df, numeric_df["c"])

        x.transform(df)


class TestInit(BaseNumericTransformerInitTests):
    """Tests for BaseNumericTransformer.init()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseNumericTransformer"


class TestFit(BaseNumericTransformerFitTests):
    """Tests for BaseNumericTransformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseNumericTransformer"


class TestTransform(BaseNumericTransformerTransformTests):
    """Tests for BaseNumericTransformer.transform()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseNumericTransformer"
